import asyncio
import logging
import time

import sys

from . import package
from .constants import MQTTv50, MQTTCommands

logger = logging.getLogger(__name__)


class _StreamReaderProtocolCompatibilityMixin:
    def __init__(self, *args, **kwargs):
        if sys.version_info < (3, 7):
            self._closed = asyncio.get_event_loop().create_future()
        super(_StreamReaderProtocolCompatibilityMixin, self).__init__(*args, **kwargs)

    def connection_lost(self, exc):
        super(_StreamReaderProtocolCompatibilityMixin, self).connection_lost(exc)

        if sys.version_info[:2] >= (3, 7):
            return

        if not self._closed.done():
            if exc is None:
                self._closed.set_result(None)
            else:
                self._closed.set_exception(exc)


class BaseMQTTProtocol(_StreamReaderProtocolCompatibilityMixin, asyncio.StreamReaderProtocol):
    def __init__(self, buffer_size=2**16, loop=None):
        if not loop:
            loop = asyncio.get_event_loop()

        self._connection = None
        self._transport = None

        self._connected = asyncio.Event()

        reader = asyncio.StreamReader(limit=buffer_size, loop=loop)
        # this is bad hack for python 3.8
        # TODO: get rid of StreamReader dependency (deprecated)
        self._hard_reader = reader
        super(BaseMQTTProtocol, self).__init__(reader, loop=loop)

    def set_connection(self, conn):
        self._connection = conn

    @property
    def closed(self):
        return self._closed

    def _parse_packet(self):
        raise NotImplementedError

    def connection_made(self, transport: asyncio.Transport):
        super(BaseMQTTProtocol, self).connection_made(transport)

        logger.info('[CONNECTION MADE]')
        self._transport = transport

        self._connected.set()

    def data_received(self, data):
        super(BaseMQTTProtocol, self).data_received(data)

    def write_data(self, data: bytes):
        self._connection._last_data_out = time.monotonic()
        if self._transport and not self._transport.is_closing():
            self._transport.write(data)
        else:
            logger.warning('[TRYING WRITE TO CLOSED SOCKET]')

    def connection_lost(self, exc):
        self._connected.clear()
        super(BaseMQTTProtocol, self).connection_lost(exc)
        if exc:
            logger.warning('[EXC: CONN LOST]', exc_info=exc)
        else:
            logger.info('[CONN CLOSE NORMALLY]')

    async def read(self, n=-1):
        bs = await self._stream_reader.read(n=n)

        # so we don't receive anything but connection is not closed -
        # let's close it manually
        if not bs and not self._transport.is_closing():
            self._transport.close()
            # self.connection_lost(ConnectionResetError())
            raise ConnectionResetError("Reset connection manually.")
        return bs


class MQTTProtocol(BaseMQTTProtocol):
    proto_name = b'MQTT'
    proto_ver = MQTTv50

    def __init__(self, *args, **kwargs):
        super(MQTTProtocol, self).__init__(*args, **kwargs)
        self._queue = asyncio.Queue()

        self._disconnect = asyncio.Event()

        self._read_loop_future = None

    def connection_made(self, transport: asyncio.Transport):
        super().connection_made(transport)
        self._read_loop_future = asyncio.ensure_future(self._read_loop())

    async def send_auth_package(self, client_id, username, password, clean_session, keepalive,
                                will_message=None, **kwargs):
        pkg = package.LoginPackageFactor.build_package(client_id, username, password, clean_session,
                                                       keepalive, self, will_message=will_message, **kwargs)
        self.write_data(pkg)

    def send_subscribe_packet(self, subscriptions, **kwargs):
        mid, pkg = package.SubscribePacket.build_package(subscriptions, self, **kwargs)
        for sub in subscriptions:
            sub.mid = mid
        self.write_data(pkg)
        return mid

    def send_unsubscribe_packet(self, topic, **kwargs):
        mid, pkg = package.UnsubscribePacket.build_package(topic, self, **kwargs)
        self.write_data(pkg)
        return mid

    def send_simple_command_packet(self, cmd):
        pkg = package.SimpleCommandPacket.build_package(cmd)
        self.write_data(pkg)

    def send_ping_request(self):
        self.send_simple_command_packet(MQTTCommands.PINGREQ)

    def send_publish(self, message):
        mid, pkg = package.PublishPacket.build_package(message, self)
        self.write_data(pkg)

        return mid, pkg

    def send_disconnect(self, reason_code=0, **properties):
        pkg = package.DisconnectPacket.build_package(self, reason_code=reason_code, **properties)

        self.write_data(pkg)

        return pkg

    def send_command_with_mid(self, cmd, mid, dup, reason_code=0):
        pkg = package.CommandWithMidPacket.build_package(cmd, mid, dup, reason_code=reason_code,
                                                         proto_ver=self.proto_ver)
        self.write_data(pkg)

    def _read_packet(self, data):
        parsed_size = 0
        raw_size = len(data)
        data_size = raw_size

        while True:
            # try to extract packet data, minimum expected packet size is 2
            if data_size < 2:
                break

            # extract payload size
            header_size = 1
            mult = 1
            payload_size = 0

            while True:
                if parsed_size + header_size >= raw_size:
                    # not full header
                    return parsed_size
                payload_byte = data[parsed_size + header_size]
                payload_size += (payload_byte & 0x7F) * mult
                if mult > 2097152:  # 128 * 128 * 128
                    return -1
                mult *= 128
                header_size += 1
                if header_size + payload_size > data_size:
                    # not enough data
                    break
                if payload_byte & 128 == 0:
                    break

            # check size once more
            if header_size + payload_size > data_size:
                # not enough data
                break

            # determine packet type
            command = data[parsed_size]
            start = parsed_size + header_size
            end = start + payload_size
            packet = data[start:end]

            data_size -= header_size + payload_size
            parsed_size += header_size + payload_size

            self._connection.put_package((command, packet))

        return parsed_size

    async def _read_loop(self):
        await self._connected.wait()

        buf = b''
        max_buff_size = 65536  # 64 * 1024
        while self._connected.is_set():
            try:
                buf += await self.read(max_buff_size)
                parsed_size = self._read_packet(buf)
                if parsed_size == -1 or self._transport.is_closing():
                    logger.debug("[RECV EMPTY] Connection will be reset automatically.")
                    break
                buf = buf[parsed_size:]
            except ConnectionResetError as exc:
                # This connection will be closed, because we received the empty data.
                # So we can safely break the while
                logger.debug("[RECV EMPTY] Connection will be reset automatically.")
                break

    def connection_lost(self, exc):
        super(MQTTProtocol, self).connection_lost(exc)
        self._connection.put_package((MQTTCommands.DISCONNECT, b''))

        if self._read_loop_future is not None:
            self._read_loop_future.cancel()
            self._read_loop_future = None

        self._queue = asyncio.Queue()



import datetime

from .client import Client, Message, Subscription
from .mqtt import constants
from .mqtt.protocol import BaseMQTTProtocol
from .mqtt.handler import MQTTConnectError

__author__ = "Mikhail Turchunovich"
__email__ = 'mitu@gurtam.com'
__copyright__ = ("Copyright 2013-%d, Gurtam; " % datetime.datetime.now().year,)

__credits__ = [
    "Mikhail Turchunovich",
    "Elena Nikolaichik"
]
__version__ = "0.6.7"


__all__ = [
    'Client',
    'Message',
    'Subscription',
    'BaseMQTTProtocol',
    'MQTTConnectError',
    'constants'
]



import asyncio
from typing import Tuple

import heapq


class BasePersistentStorage(object):
    async def push_message(self, mid, raw_package):
        raise NotImplementedError

    def push_message_nowait(self, mid, raw_package) -> asyncio.Future:
        return asyncio.ensure_future(self.push_message(mid, raw_package))

    async def pop_message(self) -> Tuple[int, bytes]:
        raise NotImplementedError

    async def remove_message_by_mid(self, mid):
        raise NotImplementedError

    @property
    async def is_empty(self) -> bool:
        raise NotImplementedError


class HeapPersistentStorage(BasePersistentStorage):
    def __init__(self, timeout):
        self._queue = []
        self._timeout = timeout

    async def push_message(self, mid, raw_package):
        tm = asyncio.get_event_loop().time()
        heapq.heappush(self._queue, (tm, mid, raw_package))

    async def pop_message(self):
        current_time = asyncio.get_event_loop().time()

        (tm, mid, raw_package) = heapq.heappop(self._queue)

        if current_time - tm > self._timeout:
            return mid, raw_package
        else:
            heapq.heappush(self._queue, (tm, mid, raw_package))

        return None

    async def remove_message_by_mid(self, mid):
        message = next(filter(lambda x: x[1] == mid, self._queue), None)
        if message:
            self._queue.remove(message)
        heapq.heapify(self._queue)

    @property
    async def is_empty(self):
        return not bool(self._queue)



import asyncio
import json

import logging
import uuid
from typing import Union, Sequence

from .mqtt.protocol import MQTTProtocol
from .mqtt.connection import MQTTConnection
from .mqtt.handler import MqttPackageHandler
from .mqtt.constants import MQTTv50, UNLIMITED_RECONNECTS

from .storage import HeapPersistentStorage

logger = logging.getLogger(__name__)


class Message:
    def __init__(self, topic, payload, qos=0, retain=False, **kwargs):
        self.topic = topic
        self.qos = qos
        self.retain = retain
        self.dup = False
        self.properties = kwargs

        if isinstance(payload, (list, tuple, dict)):
            payload = json.dumps(payload, ensure_ascii=False)

        if isinstance(payload, (int, float)):
            self.payload = str(payload).encode('ascii')
        elif isinstance(payload, str):
            self.payload = payload.encode('utf-8', errors='replace')
        elif payload is None:
            self.payload = b''
        else:
            self.payload = payload

        self.payload_size = len(self.payload)

        if self.payload_size > 268435455:
            raise ValueError('Payload too large.')


class Subscription:
    def __init__(self, topic, qos=0, no_local=False, retain_as_published=False, retain_handling_options=0,
                 subscription_identifier=None):
        self.topic = topic
        self.qos = qos
        self.no_local = no_local
        self.retain_as_published = retain_as_published
        self.retain_handling_options = retain_handling_options

        self.mid = None
        self.acknowledged = False

        # this property can be used only in MQTT5.0
        self.subscription_identifier = subscription_identifier


class Client(MqttPackageHandler):
    def __init__(self, client_id, clean_session=True, optimistic_acknowledgement=True,
                 will_message=None, **kwargs):
        super(Client, self).__init__(optimistic_acknowledgement=optimistic_acknowledgement)
        self._client_id = client_id or uuid.uuid4().hex

        # in MQTT 5.0 this is clean start flag
        self._clean_session = clean_session

        self._connection = None
        self._keepalive = 60

        self._username = None
        self._password = None

        self._host = None
        self._port = None
        self._ssl = None

        self._connect_properties = kwargs
        self._connack_properties = {}

        self._will_message = will_message

        # TODO: this constant may be moved to config
        self._retry_deliver_timeout = kwargs.pop('retry_deliver_timeout', 5)
        self._persistent_storage = kwargs.pop('persistent_storage', HeapPersistentStorage(self._retry_deliver_timeout))

        self._topic_alias_maximum = kwargs.get('topic_alias_maximum', 0)

        self._resend_task = asyncio.ensure_future(self._resend_qos_messages())

        self.subscriptions = []

    def get_subscription_by_identifier(self, subscription_identifier):
        return next((sub for sub in self.subscriptions if sub.subscription_identifier == subscription_identifier), None)

    def get_subscriptions_by_mid(self, mid):
        return [sub for sub in self.subscriptions if sub.mid == mid]

    def _remove_message_from_query(self, mid):
        logger.debug('[REMOVE MESSAGE] %s', mid)
        asyncio.ensure_future(
            self._persistent_storage.remove_message_by_mid(mid)
        )

    @property
    def is_connected(self):
        # tells if connection is alive and CONNACK was received
        return self._connected.is_set() and not self._connection.is_closing()

    async def _resend_qos_messages(self):
        await self._connected.wait()

        if await self._persistent_storage.is_empty:
            logger.debug('[QoS query IS EMPTY]')
            await asyncio.sleep(self._retry_deliver_timeout)
        elif self._connection.is_closing():
            logger.debug('[Some msg need to resend] Transport is closing, sleeping')
            await asyncio.sleep(self._retry_deliver_timeout)
        else:
            logger.debug('[Some msg need to resend] processing message')
            msg = await self._persistent_storage.pop_message()

            if msg:
                (mid, package) = msg

                try:
                    self._connection.send_package(package)
                except Exception as exc:
                    logger.error('[ERROR WHILE RESENDING] mid: %s', mid, exc_info=exc)

                await asyncio.sleep(0.001)
                await self._persistent_storage.push_message(mid, package)
            else:
                await asyncio.sleep(self._retry_deliver_timeout)

        self._resend_task = asyncio.ensure_future(self._resend_qos_messages())

    @property
    def properties(self):
        # merge two dictionaries from connect and connack packages
        return {**self._connack_properties, **self._connect_properties}

    def set_auth_credentials(self, username, password=None):
        self._username = username.encode()
        self._password = password
        if isinstance(self._password, str):
            self._password = password.encode()

    async def connect(self, host, port=1883, ssl=False, keepalive=60, version=MQTTv50, raise_exc=True):
        # Init connection
        self._host = host
        self._port = port
        self._ssl = ssl
        self._keepalive = keepalive
        self._is_active = True

        MQTTProtocol.proto_ver = version

        self._connection = await self._create_connection(
            host, port=self._port, ssl=self._ssl, clean_session=self._clean_session, keepalive=keepalive)

        await self._connection.auth(self._client_id, self._username, self._password, will_message=self._will_message,
                                    **self._connect_properties)
        await self._connected.wait()

        loop = asyncio.get_event_loop()
        while not await self._persistent_storage.is_empty:
            await loop.create_future()

        if raise_exc and self._error:
            raise self._error

    async def _create_connection(self, host, port, ssl, clean_session, keepalive):
        # important for reconnects, make sure u know what u are doing if wanna change :(
        self._exit_reconnecting_state()
        self._clear_topics_aliases()
        connection = await MQTTConnection.create_connection(host, port, ssl, clean_session, keepalive)
        connection.set_handler(self)
        return connection

    def _allow_reconnect(self):
        if self._reconnecting_now or not self._is_active:
            return False
        if self._config['reconnect_retries'] == UNLIMITED_RECONNECTS:
            return True
        if self.failed_connections <= self._config['reconnect_retries']:
            return True
        logger.error('[DISCONNECTED] max number of failed connection attempts achieved')
        return False

    async def reconnect(self, delay=False):
        if not self._allow_reconnect():
            return
        # stopping auto-reconnects during reconnect procedure is important, better do not touch :(
        self._temporatily_stop_reconnect()
        await self._disconnect()
        if delay:
            await asyncio.sleep(self._config['reconnect_delay'])
        try:
            self._connection = await self._create_connection(self._host, self._port, ssl=self._ssl,
                                                             clean_session=False, keepalive=self._keepalive)
        except OSError as exc:
            self.failed_connections += 1
            logger.warning("[CAN'T RECONNECT] %s", self.failed_connections)
            asyncio.ensure_future(self.reconnect(delay=True))
            return
        await self._connection.auth(self._client_id, self._username, self._password,
                                    will_message=self._will_message, **self._connect_properties)

    async def disconnect(self, reason_code=0, **properties):
        self._is_active = False
        self._resend_task.cancel()
        await self._disconnect(reason_code=reason_code, **properties)

    async def _disconnect(self, reason_code=0, **properties):
        self._clear_topics_aliases()

        self._connected.clear()
        if self._connection:
            self._connection.send_disconnect(reason_code=reason_code, **properties)
            await self._connection.close()

    def subscribe(self, subscription_or_topic: Union[str, Subscription, Sequence[Subscription]],
                  qos=0, no_local=False, retain_as_published=False, retain_handling_options=0, **kwargs):

        # Warn: if you will pass a few subscriptions objects, and each will be have different
        # subscription identifier - the only first will be used as identifier
        # if only you will not pass the identifier in kwargs

        subscriptions = self.update_subscriptions_with_subscription_or_topic(
            subscription_or_topic, qos, no_local, retain_as_published, retain_handling_options, kwargs)
        return self._connection.subscribe(subscriptions, **kwargs)

    def update_subscriptions_with_subscription_or_topic(
            self, subscription_or_topic, qos, no_local, retain_as_published, retain_handling_options, kwargs):

        sentinel = object()
        subscription_identifier = kwargs.get('subscription_identifier', sentinel)

        if isinstance(subscription_or_topic, Subscription):

            if subscription_identifier is not sentinel:
                subscription_or_topic.subscription_identifier = subscription_identifier

            subscriptions = [subscription_or_topic]
        elif isinstance(subscription_or_topic, (tuple, list)):

            if subscription_identifier is not sentinel:
                for sub in subscription_or_topic:
                    sub.subscription_identifier = subscription_identifier

            subscriptions = subscription_or_topic
        elif isinstance(subscription_or_topic, str):

            if subscription_identifier is sentinel:
                subscription_identifier = None

            subscriptions = [Subscription(subscription_or_topic, qos=qos, no_local=no_local,
                                          retain_as_published=retain_as_published,
                                          retain_handling_options=retain_handling_options,
                                          subscription_identifier=subscription_identifier)]
        else:
            raise ValueError('Bad subscription: must be string or Subscription or list of Subscriptions')
        self.subscriptions.extend(subscriptions)
        return subscriptions

    def resubscribe(self, subscription: Subscription, **kwargs):
        # send subscribe packet for subscription,that's already in client's subscription list
        if 'subscription_identifier' in kwargs:
            subscription.subscription_identifier = kwargs['subscription_identifier']
        elif subscription.subscription_identifier is not None:
            kwargs['subscription_identifier'] = subscription.subscription_identifier
        return self._connection.subscribe([subscription], **kwargs)

    def unsubscribe(self, topic, **kwargs):
        return self._connection.unsubscribe(topic, **kwargs)

    def publish(self, message_or_topic, payload=None, qos=0, retain=False, **kwargs):
        if isinstance(message_or_topic, Message):
            message = message_or_topic
        else:
            message = Message(message_or_topic, payload, qos=qos, retain=retain, **kwargs)

        mid, package = self._connection.publish(message)

        if qos > 0:
            self._persistent_storage.push_message_nowait(mid, package)

    def _send_simple_command(self, cmd):
        self._connection.send_simple_command(cmd)

    def _send_command_with_mid(self, cmd, mid, dup, reason_code=0):
        self._connection.send_command_with_mid(cmd, mid, dup, reason_code=reason_code)

    @property
    def protocol_version(self):
        return self._connection._protocol.proto_ver \
            if self._connection is not None else MQTTv50
